import os
import pandas as pd
import xarray as xr
import numpy as np
from scipy.spatial import cKDTree
from datetime import date, timedelta, datetime
from mpi4py import MPI
import json
import vars_to_int as vti

class NetCDFInterpolator:
    def __init__(self, expid, start_d, end_d, stn_lat, stn_lon):
        self.expid = expid
        self.start_d = start_d
        self.end_d = end_d
        self.stn_lat = stn_lat
        self.stn_lon = stn_lon
        self.n_neighbours = 4
        self.base_dir = f'/esarchive/exp/monarch/{self.expid}/'
        self.doms = [i for i in os.listdir(self.base_dir) if i.startswith("d")]
        self.bool_check = any( i in self.doms for i in os.listdir(self.base_dir))
        self.data = None

    def _get_folder_path(self, variable):
        if self.bool_check:
            return f'{self.base_dir}{dom}/hourly/{variable}/'
        else:
            return f'{self.base_dir}regional/hourly/{variable}/'

    def read_netcdfs(self, variables):
        dfs = []
        for variable in variables:
            folder_path = self._get_folder_path(variable)
            if os.path.isdir(folder_path):
                dates = [datetime.strftime(self.start_d + timedelta(days=i), '%Y%m%d') for i in range((self.end_d - self.start_d).days + 1)]
                list_of_nc_files = [f'{folder_path}{variable}-000_{i}00.nc' for i in dates]
                df = self._process_variable(list_of_nc_files, variable)
                dfs.append(df)
            else:
                print(f"No se encontró la ruta para la variable '{variable}'.")
        self.data = pd.concat(dfs, axis=1)

    def lon_lat_to_cartesian(self, lon, lat):
        A = 6378.137  # major axis [km]
        E2 = 6.69437999014e-3  # eccentricity squared
        # Verifica si 'lon' es de tipo numérico
        try:
            lon = lon.astype(float)
            lat = lat.astype(float)
            lon_rad = np.radians(lon)
            lat_rad = np.radians(lat)
        except ValueError:
            print("Error: 'lon' contiene valores que no pueden ser convertidos a flotantes.")

        #lon_rad = np.radians(lon)
        #lat_rad = np.radians(lat)
        r_n = A / (np.sqrt(1 - E2 * (np.sin(lat_rad) ** 2)))
        x = r_n * np.cos(lat_rad) * np.cos(lon_rad)
        y = r_n * np.cos(lat_rad) * np.sin(lon_rad)
        z = r_n * (1 - E2) * np.sin(lat_rad)
        return x, y, z

    def interpolate(self, source, target, dist, inds, variable):
        tmp = []
        for t in range(source["time"].shape[0]):
            w = 1.0 / dist ** 2
            interp_values = np.sum(w * source[variable][t].values.flatten()[inds], axis=1) / np.sum(w, axis=1)
            interp_values.shape = target.shape
            tmp.append(interp_values)
        return tmp

    def save_data(self, output_file):
        if self.data is not None:
            self.data.to_csv(output_file, index=True)
        else:
            print("No hay datos para almacenar. Por favor, lee los archivos NetCDF primero.")

    def _process_variable(self, file_list, variable):
        dataframes = []
        for file_name in file_list:
            file_path = os.path.join(self._get_folder_path(variable), file_name)
            if os.path.isfile(file_path):
                with xr.open_dataset(file_path) as ds:
                    model_files = ds.isel(time=slice(None, -1))
                    lon_target, lat_target = np.meshgrid(self.stn_lon, self.stn_lat)
                    lat_source, lon_source = model_files.lat, model_files.lon
                    xs, ys, zs = self.lon_lat_to_cartesian(lon_source.values.flatten(), lat_source.values.flatten())
                    xt, yt, zt = self.lon_lat_to_cartesian(lon_target.flatten(), lat_target.flatten())
                    tree = cKDTree(np.column_stack((xs, ys, zs)))
                    dist, inds = tree.query(np.column_stack((xt, yt, zt)), k=self.n_neighbours)
                    interp_vals = self.interpolate(model_files, lon_target, dist, inds, variable)
                    if model_files[variable].attrs.get('units') == 'kgm-3':
                        x = [interp_val[0][0] * 1e9 for interp_val in interp_vals]  # Convert kg/m3 to ug/m3
                    else:
                        x = [interp_val[0][0] for interp_val in interp_vals]
                    times = model_files['time'].values
                    df = pd.DataFrame({variable: x}, index=times)
                    dataframes.append(df)
            else:
                print(f"No se encontró el archivo '{file_path}'.")
        return pd.concat(dataframes)

    #building the output_folder
    def output_dir(self):
        # Construct the directory path
        directory_path = os.path.join('/esarchive/scratch/hnavarro/int_expids/', self.expid, str(self.start_d.year))

        # Create the directory if it doesn't exist
        if not os.path.isdir(directory_path):
            os.makedirs(directory_path, exist_ok=True)

        return directory_path


# Inicializa MPI
comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

# Carga las estaciones desde el archivo JSON
with open("nstations.json", "r") as f:
    stations_dict = json.load(f)

# Convierte el diccionario a una lista de tuples (nombre, lat, lon)
stations = [(name, data['lat'], data['lon']) for name, data in stations_dict.items()]

# Divide las estaciones entre los procesos
stations_split = np.array_split(stations, size)
variables = vti.brc_vars #vti.vars_bc_abs #vti.vars_to_tag 

expid = "a5lj"
start_d = date(2018,1,1)
end_d   = date(2018,12,31)
# Cada proceso procesa su porción de estaciones
for station in stations_split[rank]:
    name, lat, lon = station
    reader = NetCDFInterpolator(expid, start_d, end_d, lat, lon)
    reader.read_netcdfs(variables)  # Asegúrate de tener una lista de las variables que deseas leer
    output_dir = reader.output_dir()
    output_file = os.path.join(output_dir, f'{expid}_{name}.csv')
    reader.save_data(output_file)
